import os, sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "")))

import json
import re
import urllib.request

import utils.constants
import utils.request

URL_V1 = utils.constants.URL_V1
URL_V2 = utils.constants.URL_V2
DEFAULT_TIMEOUT = utils.constants.DEFAULT_TIMEOUT
INGREDIENT_QUERIES = utils.constants.INGREDIENT_QUERIES
SKILLS = utils.constants.SKILLS
SPRITES = utils.constants.SPRITES
IDENTIFICATIONS = utils.constants.IDENTIFICATIONS
ITEM_ONLY_IDS = utils.constants.ITEM_ONLY_IDS
CONSUMABLE_ONLY_IDS = utils.constants.CONSUMABLE_ONLY_IDS
CATEGORIES = utils.constants.CATEGORIES
RECIPE_CATEGORIES = utils.constants.RECIPE_CATEGORIES
RECIPE_QUERIES = utils.constants.RECIPE_QUERIES
RECIPE_MIN_MAX = utils.constants.RECIPE_MIN_MAX


class Guild:
    # https://docs.wynncraft.com/Guild-API/

    url = URL_V1 + "guild"
    
    def list():
        url = Guild.url + "List"
        return utils.request.open(url)
    
    def stats(guild_name):
        url = Guild.url + f"Stats&command={guild_name}"
        return utils.request.open(url)


class Ingredient:
    # https://docs.wynncraft.com/Ingredient-API/

    url = URL_V2 + "ingredient/"

    def get(name):
        url = Ingredient.url + f"get/{name}"
        return utils.request.open(url)

    def list():
        url = Ingredient.url + "list/"
        return utils.request.open(url)
    
    def search(query, arg):
        url = Ingredient.url + "search/"

        raise_error = False

        if query not in INGREDIENT_QUERIES:
            raise TypeError(f"search() invaild query: '{query}'")
        
        if query == "name":
            pass

        elif query == "tier":
            if not (0 <= int(arg) <= 3):
                raise_error = True

        elif query == "level":
            if int(arg) < 0:
                raise_error = True

        elif query == "skills":
            re.IGNORECASE = True
            regex = re.compile(f"^[&^]({'|'.join(SKILLS)})(,({'|'.join(SKILLS)}))*$")
            if not re.fullmatch(regex, arg):
                raise_error = True

        elif query == "sprite":
            re.IGNORECASE = False
            regex = re.compile(f"^[&^](({'|'.join(SPRITES)})<\d+>)(,(({'|'.join(SPRITES)})<\d+>))*$")
            if not re.fullmatch(regex, arg):
                raise_error = True

        elif query == "identifications":
            re.IGNORECASE = True
            regex = re.compile(f"^[&^](({'|'.join(IDENTIFICATIONS)})<\d*;\d*>)(,(({'|'.join(IDENTIFICATIONS)})<\d*;\d*>))*$")
            if not re.fullmatch(regex, arg):
                raise_error = True

        elif query == "itemOnlyIDs":
            re.IGNORECASE = False
            regex = re.compile(f"^[&^](({'|'.join(ITEM_ONLY_IDS)})<\d+>)(,(({'|'.join(ITEM_ONLY_IDS)})<\d+>))*$")
            if not re.fullmatch(regex, arg):
                raise_error = True

        elif query == "consumableOnlyIDs":
            re.IGNORECASE = False
            regex = re.compile(f"^[&^](({'|'.join(CONSUMABLE_ONLY_IDS)})<\d+>)(,(({'|'.join(CONSUMABLE_ONLY_IDS)})<\d+>))*$")
            if not re.fullmatch(regex, arg):
                raise_error = True

        else:
            pass
        
        if raise_error:
            raise TypeError(f"search() invaild argument for {query} query: {arg}")

        url += f"{query}/{arg}"
        return utils.request.open(url)
    
    def search_name(arg):
        return Ingredient.search("name", arg)

    def search_tier(arg):
        return Ingredient.search("tier", arg)

    def search_level(arg):
        return Ingredient.search("level", arg)

    def search_skills(arg):
        return Ingredient.search("skills", arg)

    def search_sprite(arg):
        return Ingredient.search("sprite", arg)

    def search_identifications(arg):
        return Ingredient.search("identifications", arg)

    def search_item_only_ids(arg):
        return Ingredient.search("itemOnlyIDs", arg)

    def search_consumable_only_ids(arg):
        return Ingredient.search("consumableOnlyIDs", arg)

class Item:
    # https://docs.wynncraft.com/Item-API/

    url = URL_V1 + "item"
    
    def database_category(category):
        if category not in CATEGORIES:
            raise TypeError(f"database_category() invaild category: {category}")

        url = Item.url + f"DB&category={category}"
        return utils.request.open(url)
    
    def database_search(name):
        url = Item.url + f"DB&search={name}"
        return utils.request.open(url)


class Leaderboard:
    # https://docs.wynncraft.com/Leaderboard-API

    url = URL_V1 + "statsLeaderboard"

    def guild(timeframe):
        url = Leaderboard.url + f"&type=guild&timeframe={timeframe}"
        return utils.request.open(url)
    
    def player(timeframe):
        url = Leaderboard.url + f"&type=player&timeframe={timeframe}"
        return utils.request.open(url)

    def pvp(timeframe):
        url = Leaderboard.url + f"&type=pvp&timeframe={timeframe}"
        return utils.request.open(url)


class Network:
    # https://docs.wynncraft.com/Network-API/

    url = URL_V1 + "onlinePlayers"

    def server_list():
        url = Network.url
        return utils.request.open(url)

    def player_sum():
        url = Network.url + "Sum"
        return utils.request.open(url)


class Player:
    # https://docs.wynncraft.com/Player-API/

    url = URL_V2 + "player/"

    def stats(player):
        url = Player.url + f"{player}/stats"
        return utils.request.open(url)

    def uuid(username):
        url = Player.url + f"{username}/uuid"
        return utils.request.open(url)


class Recipe:
    # https://docs.wynncraft.com/Recipe-API/

    url = URL_V2 + "recipe/"
    
    def get(name):
        for idx, word in enumerate(RECIPE_CATEGORIES):
            RECIPE_CATEGORIES[idx] = word.capitalize()
        
        re.IGNORECASE = False
        regex = re.compile(f"^({'|'.join(RECIPE_CATEGORIES)})-\d+-\d+$")
        if not re.fullmatch(regex, name):
            raise TypeError(f"get() invaild name: {name}")

        url = Recipe.url + f"get/{name}"
        return utils.request.open(url)
    
    def list():
        url = Recipe.url + "/list"
        return utils.request.open(url)

    def search(query, arg):
        url = Recipe.url + "search/"

        raise_error = False

        if query not in RECIPE_QUERIES:
            raise TypeError(f"search() invaild query: '{query}'")

        if query == "type":
            pass

        elif query == "skill":
            re.IGNORECASE = True
            regex = re.compile(f"({'|'.join(SKILLS)})")
            if not re.fullmatch(regex, arg):
                raise_error = True
        
        elif query in ["level", "durability", "healthOrDamage", "duration", "basicDuration"]:
            re.IGNORECASE = False
            regex = re.compile(f"^[&^](({'|'.join(RECIPE_MIN_MAX)})<\d+>)(,(({'|'.join(RECIPE_MIN_MAX)})<\d+>))*$")
            if not re.fullmatch(regex, arg):
                raise_error = True
        
        else:
            pass

        if raise_error:
            raise TypeError(f"search() invaild argument for {query} query: {arg}")

        url += f"{query}/{arg}"
        return utils.request.open(url)

    def search_type(arg):
        return Recipe.search("type", arg)

    def search_skill(arg):
        return Recipe.search("skill", arg)

    def search_level(arg):
        return Recipe.search("level", arg)
    
    def search_durability(arg):
        return Recipe.search("durability", arg)

    def search_healthOrDamage(arg):
        return Recipe.search("healthOrDamage", arg)

    def search_duration(arg):
        return Recipe.search("duration", arg)

    def search_basicDuration(arg):
        return Recipe.search("basicDuration", arg)


class Search:
    # https://docs.wynncraft.com/Search-API/

    url = URL_V1 + "statsSearch"

    def name(name):
        url = Search.url + f"&search={name}"
        return utils.request.open(url)


class Territory:
    # https://docs.wynncraft.com/Territory-API/

    url = URL_V1 + "territoryList"

    def list():
        url = Territory.url
        return utils.request.open(url)



from wynncraft.cache.cache import Guild
from wynncraft.cache.cache import Item
from wynncraft.cache.cache import Ingredient
from wynncraft.cache.cache import Leaderboard
from wynncraft.cache.cache import Network
from wynncraft.cache.cache import Player
from wynncraft.cache.cache import Recipe
from wynncraft.cache.cache import Search
from wynncraft.cache.cache import Territory



# Constants for the wrapper
"""
Specifies a timeout in seconds for http request.
Default: 10
"""
DEFAULT_TIMEOUT = 10

"""
If the data in cache is older than CACHE_TIME seconds, then a new request will be made.
Default: 300
"""
CACHE_TIME = 300


# Constants for communicating Wynncraft API (DO NOT CHANGE!)
URL_V1 = "https://api.wynncraft.com/public_api.php?action="

URL_V2 = "https://api.wynncraft.com/v2/"

URL_CODES = {
    " ": "+",
    "^": "%5E",
    "<": "%3C",
    ">": "%3E"
}

INGREDIENT_QUERIES =[
    "name",
    "tier",
    "level",
    "skills",
    "sprite",
    "identifications",
    "itemOnlyIDs",
    "consumableOnlyIDs"
]

SKILLS = [
    "alchemism",
    "armouring",
    "cooking",
    "jeweling",
    "scribing",
    "tailoring",
    "weaponsmithing",
    "woodworking"
]

SPRITES = [
    "id",
    "damage"
]

IDENTIFICATIONS = [
    "agilitypoints",
    "airdamagebonus",
    "airdefense",
    "attackspeed",
    "damagebonus",
    "damagebonusraw",
    "defensepoints",
    "dexteritypoints",
    "earthdamagebonus",
    "earthdefense",
    "emeraldstealing",
    "exploding",
    "firedamagebonus",
    "firedefense",
    "healthbonus",
    "healthregen",
    "healthregenraw",
    "intelligencepoints",
    "lifesteal",
    "lootbonus",
    "loot_quality",
    "manaregen",
    "manasteal",
    "poison",
    "reflection",
    "soulpoints",
    "speed",
    "spelldamage",
    "spelldamageraw",
    "stamina_regen",
    "strengthpoints",
    "thorns",
    "thunderdamagebonus",
    "thunderdefense",
    "waterdamagebonus",
    "waterdefense",
    "xpbonus"
]

ITEM_ONLY_IDS = [
    "durability",
    "strength",
    "dexterity",
    "intelligence",
    "defence",
    "agility"
]

CONSUMABLE_ONLY_IDS = [
    "duration",
    "charges"
]

CATEGORIES = [
    "all",
    "boots",
    "bow",
    "bracelet",
    "chestplate",
    "dagger",
    "helmet",
    "leggings",
    "necklace",
    "ring",
    "spear", 
    "wand"
]

RECIPE_CATEGORIES = [
    "boots",
    "bow",
    "bracelet",
    "chestplate",
    "dagger",
    "food",
    "helmet",
    "leggings",
    "necklace",
    "potion",
    "relik",
    "ring",
    "scroll",
    "spear", 
    "wand"
]

RECIPE_QUERIES = [
    "type",
    "skill",
    "level",
    "durability",
    "healthOrDamage",
    "duration",
    "basicDuration"
]

RECIPE_MIN_MAX = [
    "min",
    "max"
]



import os, sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

import json
import time

import utils.constants
import wynncraft


def run_cache(id, function, *args):
    if CacheManager.call_request(id):
        exec(f"x = {function}{args}", globals(), globals())
        res = x
    else:
        res = None

    return CacheManager.search_cache(id, res)


class CacheManager:
    try:
        os.chdir(os.path.abspath(os.getcwd()) + "/.cache")
    except FileNotFoundError:
        os.mkdir(os.path.abspath(os.getcwd()) + "/.cache")
        os.chdir(os.path.abspath(os.getcwd()) + "/.cache")

    def read_cache():
        try:
            open(".cache.json")
        except FileNotFoundError:
            open(".cache.json", "w")
            return json.loads("{}")
        else:
            with open(".cache.json", "r") as c:
                try:
                    return json.loads(c.read())
                except json.decoder.JSONDecodeError:
                    return json.loads("{}")


    def write_cache(data):
        cache = {}
        try:
            cache = json.loads(open(".cache.json").read())
        except (FileNotFoundError,json.decoder.JSONDecodeError):
            cache = {}
        finally:
            with open(".cache.json", "w") as c:
                cache.update(data)
                json.dump(cache, c)

    def read_cache_table():
        try:
            open(".cache-table.json")
        except FileNotFoundError:
            open(".cache-table.json", "w")
            return json.loads("{}")
        else:
            with open(".cache-table.json", "r") as c:
                try:
                    return json.loads(c.read())
                except json.decoder.JSONDecodeError:
                    return json.loads("{}")

    def write_cache_table(data):
        cache = {}
        try:
            cache = json.loads(open(".cache-table.json").read())
        except (FileNotFoundError,json.decoder.JSONDecodeError):
            cache = {}
        finally:
            with open(".cache-table.json", "w") as c:
                cache.update(data)
                json.dump(cache, c)

    def delete_cache():
        for f in [".cache.json", ".cache-table.json"]:
            try:
                os.remove(f)
            except FileNotFoundError:
                continue

    def search_cache(id, res):
        if res:
            CacheManager.write_cache({id: res})
            CacheManager.write_cache_table({id: int(time.time()) + utils.constants.CACHE_TIME})
            return res
        else:
            return CacheManager.read_cache()[id]
    
    def call_request(id):
        cache_table = CacheManager.read_cache_table()
        return ((not cache_table) or (id not in cache_table) or (cache_table[id] < int(time.time())))

    
class Guild:
    def list():
        return run_cache("guild_list", "wynncraft.Guild.list")

    def stats(name):
        return run_cache("guild_stats", "wynncraft.Guild.stats", name)


class Ingredient:
    def get(name):
        return run_cache(f"ingredient_get_{name}", "wynncraft.Ingredient.get", name)

    def list():
        return run_cache(f"ingredient_list", "wynncraft.Ingredient.list")

    def search(query, arg):
        return run_cache(f"ingredient_search_{query}_{arg}", "wynncraft.Ingredient.search", query, arg)

    def search_name(arg):
        return Ingredient.search("name", arg)
    
    def search_tier(arg):
        return Ingredient.search("tier", arg)

    def search_level(arg):
        return Ingredient.search("level", arg)

    def search_skills(arg):
        return Ingredient.search("skills", arg)

    def search_sprite(arg):
        return Ingredient.search("sprite", arg)
    
    def search_identifications(arg):
        return Ingredient.search("identifications", arg)

    def search_item_only_ids(arg):
        return Ingredient.search("itemOnlyIDs", arg)

    def search_consumable_only_ids(arg):
        return Ingredient.search("consumableOnlyIDs", arg)


class Item:
    def database_category(category):
        return run_cache(f"item_db_category_{category}", "wynncraft.Item.database_category", category)

    def database_search(name):
        return run_cache(f"item_db_search_{name}", "wynncraft.Item.database_search", name)


class Leaderboard:
    def guild(timeframe):
        return run_cache(f"leaderboard_guild_{timeframe}", "wynncraft.Leaderboard.guild", timeframe)

    def player(timeframe):
        return run_cache(f"leaderboard_player_{timeframe}", "wynncraft.Leaderboard.player", timeframe)

    def pvp(timeframe):
        return run_cache(f"leaderboard_pvp_{timeframe}", "wynncraft.Leaderboard.pvp", timeframe)


class Network:
    def server_list():
        return run_cache("server_list", "wynncraft.Network.server_list")

    def player_sum():
        return run_cache("player_sum", "wynncraft.Network.player_sum")


class Player:
    def stats(player):
        return run_cache("player_stats", "wynncraft.Player.stats", player)

    def uuid(username):
        return run_cache("player_uuid", "wynncraft.Player.uuid", username)


class Recipe:
    def get(name):
        return run_cache(f"recipe_get_{name}", "wynncraft.Recipe.get", name)

    def list():
        return run_cache("recipe_list", "wynncraft.Recipe.list")

    def search(query, arg):
        return run_cache(f"recipe_search_{query}_{arg}", "wynncraft.Recipe.search", query, arg)

    def search_type(arg):
        return Recipe.search("type", arg)
    
    def search_skill(arg):
        return Recipe.search("skill", arg)

    def search_level(arg):
        return Recipe.search("level", arg)
    
    def search_durability(arg):
        return Recipe.search("durability", arg)

    def search_healthOrDamage(arg):
        return Recipe.search("healthOrDamage", arg)

    def search_duration(arg):
        return Recipe.search("duration", arg)

    def search_basicDuration(arg):
        return Recipe.search("basicDuration", arg)


class Search:
    def name(name):
        return run_cache(f"search_{name}", "wynncraft.Search.name", name)


class Territory:
    def list():
        return run_cache("territory_list", "wynncraft.Territory.list")



import json
import urllib.request

import utils.constants


def open(url):
    for char in url:
        if char in utils.constants.URL_CODES:
            url = url.replace(char, utils.constants.URL_CODES[char])
    
    return json.loads(urllib.request.urlopen(url, timeout=utils.constants.DEFAULT_TIMEOUT).read().decode("utf-8"))



import wynncraft.cache

from wynncraft.wynncraft import Guild
from wynncraft.wynncraft import Item
from wynncraft.wynncraft import Ingredient
from wynncraft.wynncraft import Leaderboard
from wynncraft.wynncraft import Network
from wynncraft.wynncraft import Player
from wynncraft.wynncraft import Recipe
from wynncraft.wynncraft import Search
from wynncraft.wynncraft import Territory
